'''
Created on 28 Jul 2017

@author: julianporter
'''

import random
import numpy as np
from OSGridConverter.cartesian import Vector, Matrix
from OSGridConverter import LatLong, OSGridReference
from math import log10

def equalVectors(np,v):
    if len(np)!=len(v): return False
    return all([np[i]==v[i] for i in range(len(v))])

def equalMatrices(np,v):
    if len(np)!=len(v): return False
    return all([equalVectors(np[i],v[i]) for i in range(len(v))])

def vectorError(np,v):
    return sum([abs(np[i]-v[i]) for i in range(len(v))])

def vectorDiff(np,v):
    e=vectorError(np,v)
    if e==0 : return 1000.0
    return -log10(e)

def matrixDiff(np,v):
    e=sum([vectorError(np[i],v[i]) for i in range(len(v))])
    if e==0 : return 1000.0
    return -log10(e)

def gridDiff(g1,g2):
    e=max([abs(x-y) for (x,y) in zip(g1.eastNorth(),g2.eastNorth())])
    if e==0 : return 1000.0
    return 6-log10(e)

def latlongDiff(l1,l2):
    e=(l1.latitude-l2.latitude,l1.longitude-l2.longitude)
    e=max([abs(x) for x in e])
    if e==0 : return 1000.0
    return 3-log10(e)

class Generator(object):
    
    def __init__(self,lo=-100.0,hi=100.0,n=3):
        self.lo=lo
        self.hi=hi
        self.n=n
        
    def randomScalar(self):
        return random.uniform(self.lo,self.hi)
        
    def randomList(self):
        return [random.uniform(self.lo,self.hi) for _ in range(self.n)]
    
    def randomListOfLists(self):
        return [self.randomList() for _ in range(self.n)]
    
    def randomNPVector(self):
        return np.array(self.randomList())
    
    def randomNPMatrix(self):
        return np.array(self.randomListOfLists())
    
    def randomVector(self):
        return Vector(self.randomList(),n=self.n)
    
    def randomMatrix(self):
        return Matrix(self.randomListOfLists(),n=self.n)
    
    def randomVectorPair(self):
        a=self.randomList()
        return (np.array(a), Vector(a,n=self.n))
    
    def randomMatrixPair(self):
        a=self.randomListOfLists()
        return (np.array(a), Matrix(a,n=self.n))
    
    def randomLatLong(self,tag='WGS84'):
        latitude  = random.uniform(50.0,59.0)
        longitude = random.uniform(-5.0,2.0)
        return LatLong(latitude,longitude,tag=tag)
    
    def randomGrid(self,tag='WGS84'):
        ll=self.randomLatLong(tag=tag)
        return OSGridReference(ll)
    




'''
Created on 27 Jul 2017

@author: julianporter
'''

from math import sqrt
from ..base import OSGridError, areNumbers

def dot(a1,a2):
    zipped=zip(a1,a2)
    mixed=map(lambda p : p[0]*p[1],zipped)
    return sum(mixed)



class Vector(object):
    
    @classmethod
    def zero(cls,n=3):
        return Vector([0]*n,n=n)
        
    def __init__(self,args,n=3):
        self.n=n
        if len(args)!=n : raise OSGridError('Constructor of Vector takes {} arguments'.format(n))
        if isinstance(args,Vector):
            self.array=args.array
        elif areNumbers(args):
            self.array=args
        else:
            raise OSGridError('Constructor of Vector rerquires Vector or [numeric] arguments')
    
    def __len__(self):
        return self.n
       
    def __iter__(self):
        return iter(self.array)
    
    def __getitem__(self,key):
        return self.array[key]

    def _i(self,array):
        return Vector(array,n=self.n)
    
    def __add__(self,other):
        return self._i([x+y for (x,y) in zip(self,other)])
    
    def __sub__(self,other):
        return self._i([x-y for (x,y) in zip(self,other)])
    
    def __neg__(self):
        return self._i([- x for x in self.array])
    
    def __mul__(self,other):
        return self._i([x*other for x in self.array])
    
    def __rmul__(self,other):
        return self._i([x*other for x in self.array])
    
    def __or__(self,other):
        return sum([x*y for (x,y) in zip(self,other)])
    
    def __eq__(self,other):
        return all([x==y for (x,y) in zip(self,other)])
    
    def __str__(self):
        return str(self.array)
    
    
class Vector3(Vector):
    
    def __init__(self,args):
        super(Vector3,self).__init__(args,n=3)
        self.x, self.y, self.z = self.array
        
    @property
    def p(self):
        return sqrt(self.x*self.x + self.y*self.y)
    
    @property
    def R(self):
        return self.p*self.p+self.z*self.z
            
class Matrix(object):
    
    @classmethod
    def outerProduct(cls,left,right=None,n=3):
        r= right or left
        if len(left)!=n or len(r)!=n:
            raise OSGridError('Outer product requires two vectors of length {}'.format(n))
        rows=[r*rv for rv in left]
        return Matrix(rows,n)
    
    @classmethod
    def identity(cls,n=3):
        rows=[]
        for i in range(n):
            r=[0]*n
            r[i]=1
            rows.append(r)
        return Matrix(rows,n=n)
        
    
    def __init__(self,rows,n=3):
        self.n=n
        if len(rows)!=n or any([len(r)!=3 for r in rows]):
            raise OSGridError('Constructor of Matrix takes {} rows of {} numeric arguments'.format(n,n))
            
        self.rows=[Vector(row,n=self.n) for row in rows]
        self.columns=[Vector([row[index] for row in self.rows],n=self.n) for index in range(self.n)]
            
    def row(self,index):
        return self.rows[index]
    
    def column(self,index):
        return self.columns[index]
    
    def __getitem__(self,index):
        return self.rows[index]
    
    def __len__(self):
        return self.n
    
    def __abs__(self):
        return sum([x*x for x in self])
    
    def __add__(self,other):
        if len(self)!=len(other):
            raise OSGridError('Sum requires elements of same size')
        rows=[x+y for (x,y) in zip(self.rows,other.rows)]
        return Matrix(rows)

    def __matmul__(self,other):
        if len(self)!=len(other):
            raise OSGridError('Product requires elements of same size')
        
        if isinstance(other,Vector):
            row=[row|other for row in self.rows]
            return Vector(row,n=self.n)
        elif isinstance(other,Matrix):
            rows=[self @ column for column in other.columns]
            return Matrix(rows,n=self.n)
        else:
            raise OSGridError('Product requires Vector or Matrix')
        
    def __mul__(self,other):
        return Matrix([r*other for r in self.rows],n=3)
    
    def __rmul__(self,other):
        return Matrix([other*r for r in self.rows],n=3)
        
    def __str__(self):
        return '[{}]'.format(','.join([str(row) for row in self.rows]))
    
    def __eq__(self,other):
        return all([self.row[i]==other.row[i] for i in range(self.n)])
                
                
            
    
    


'''
Created on 27 Jul 2017

@author: julianporter
'''

from math import radians
from .algebra import Vector, Matrix
from ..base import OSGridError, areNumbers

def negate(a):
    return [-x for x in a]

class HelmertMatrix(Matrix):
    
    @classmethod
    def rowsFor(cls,r,s):
        return [[ s  , -r[2],  r[1]],
                [ r[2], s   , -r[0]],
                [-r[1], r[0],  s]]
    
    def __init__(self,*args):
        if len(args)==0: raise OSGridError('HelmertMatrix requires at least one argument')
        elif isinstance(args[0],Matrix):
            super(HelmertMatrix,self).__init__(args[0].rows,n=3)
        elif len(args)==2 and isinstance(args[0],Vector) and areNumbers([args[1]]):
            self.r=args[0]
            self.s=1.0+args[1]
            super(HelmertMatrix,self).__init__(self.rowsFor(self.r,self.s),n=3)
        else:
            raise OSGridError('HelmertMatrix requires matrix or vector + scalar')
    
    def determinant(self):
        return self.s*((self.r|self.r)+self.s*self.s)
    
    def inverse(self):
        m=(self.s*Matrix(self.rowsFor(-self.r,self.s),n=3))+Matrix.outerProduct(self.r, n=3)
        return m*(1.0/self.determinant())      
         

class Transform (object) :
    


    @classmethod
    def _normalise(cls,seconds):
        return radians(seconds/3600.0)
    
    def __init__(self,*args):
        if len(args)>=3:
            t, r, s = args
            self.t=Vector(t)
            
            r=Vector([radians(x/3600.0) for x in r])
            s=s/1.0e6
            self.mx=HelmertMatrix(r,s)
        elif len(args)==0:
            self.mx=Matrix.identity(3)
            self.t=Vector.zero(3)
        
    def __call__(self,vector):
        return self.t + (self.mx @ vector)
        
    @property
    def inverse(self):
        inv=Transform()
        inv.mx=self.mx.inverse()
        inv.t =-(inv.mx@self.t)
        return inv
    
    def __str__(self):
        return 'mx = {}, t = {}'.format(self.mx,self.vecT)

    
    
        



'''
Created on 27 Jul 2017

@author: julianporter
'''

from math import sqrt, isnan, cos, sin, atan2, degrees, radians
from ..base import OSGridError,log
from .algebra import Vector,Vector3

class Cartesian (Vector3):
    def __init__(self,*args):
        if len(args)==0: raise OSGridError('Initialisation of Cartesian requires at least one argument')
        if isinstance(args[0],Vector):
            super(Cartesian,self).__init__(args[0].array)
        else:
            try:
                latlong=args[0]
                phi = radians(latlong.latitude)
                l   = radians(latlong.longitude)
                
                a=latlong.datum.ellipsoid.a
                f=latlong.datum.ellipsoid.f
                
                s=sin(phi)
                c=cos(phi)
        
                eSq=2*f-f*f
                nu=a/sqrt(1.0-eSq*s*s)
                
                super(Cartesian,self).__init__([nu*c*cos(l),nu*c*sin(l),nu*(1-eSq)*s])
            except:
                super(Cartesian,self).__init__(args)
        
    def LatLong(self,ellipsoid):
        
        a = ellipsoid.a
        b = ellipsoid.b
        f = ellipsoid.f

        e1  = 2*f - f*f
        e2  = e1/(1-e1)
    

        t = (1+e2*b/self.R)*b*self.z/(a*self.p) 
        s = t/sqrt(1+t*t)
        c = s/t
        
        if isnan(c): phi=0
        else: phi=atan2(self.z + e2*b*s*s*s, self.p - e1*a*c*c*c) 
        l = atan2(self.y, self.x)

        return (degrees(phi),degrees(l))
    
    def transform(self,transform):
        transformed=transform(self)
        log.debug('Transform of {} is {}',self,transformed)
        return Cartesian(*transformed.array)
    
    
    

    def __str__(self):
        return '({},{},{})'.format(self.x,self.y,self.z)
     
       
        


'''
Created on 26 Jul 2017

@author: julianporter
'''

import unittest
from .baseTest import TestFramework
from OSGridConverter import OSGridReference
from .types import gridDiff,latlongDiff


class GridToLatLongAndBack(TestFramework):
    
    testName='grid to lat/long to grid'
    
    def setUp(self):
        super(GridToLatLongAndBack,self).setUp()
        self.tag='WGS84'
    
    def action(self):
        grid1=self.generator.randomGrid(tag=self.tag)
        if len(str(grid1))>0:
            ll=grid1.toLatLong(tag=self.tag, viaString=True)
            grid2=OSGridReference(ll)
            error=gridDiff(grid1,grid2)
            self.push(error,3)
            er=(str(grid1)==str(grid2)) or (error>3.0)
            if er: return 1
            else:
                print('{}  {} -  {}  {}  : {} {}'.format(grid1,grid1.eastNorth(),grid2,grid2.eastNorth(),error,er))
                return 0


                
   

class LatLongToGridAndBack (TestFramework):
    
    testName='lat/long to grid to lat/long'
    
    def setUp(self):
        super(LatLongToGridAndBack,self).setUp()
        self.tag='WGS84'
        self.error=10
            
    def action(self):
        ll1=self.generator.randomLatLong(tag=self.tag)
        grid=OSGridReference(ll1)
        ll2=grid.toLatLong(tag=self.tag)
        #e=(ll1.latitude-ll2.latitude,ll1.longitude-ll2.longitude)
        error=latlongDiff(ll1,ll2)
        #print('{} - {} - {} : {} {}'.format(ll1,grid,ll2,e,error))
        return  self.push(error,3)
        
        
        
        
    
 
if __name__ == '__main__':
    print('Running 100 lat/long -> National Grid -> lat / long conversions')
    unittest.main()



'''
Created on 19 Jul 2017

@author: julianporter
'''


import re
from math import radians, degrees , pow, floor
from numbers import Number
from .mapping import OSDefaultGrid
from .latlong import LatLong
from .base import OSGridError, areNumbers, log
import traceback

        
    

def areNumbers(args):
    return all([isinstance(x,Number) and not isinstance(x,bool) for x in args])

class OSGridReference (object):
    
    alphabet='ABCDEFGHJKLMNOPQRSTUVWXYZ'
    
 
    def __init__(self,*args):
        if len(args)==0: raise OSGridError('Initialisation of OSGRidReference requires at least one argument')
        arg=args[0]
        
        if isinstance(arg,str):
            log.debug('Parsing {}',arg)
            gridref = arg.strip().upper()

            log.debug('gridref is {}',gridref)
            match = re.match('^(\d+),\s*(\d+)$',gridref)
            if match:
                log.debug('Matched pair of numbers')
                return [int(x) for x in match.groups()]
        

            match = re.match('^([A-Z]{2})\s*([0-9]+)(\s*)([0-9]+)$',gridref)
            if not match:
                log.error("Couldn't match expected format")
                raise RuntimeError('Invalid grid reference')
        
            g=match.groups()
            alpha=g[0]
            l1 = ord(alpha[0]) - ord('A')
            l2 = ord(alpha[1]) - ord('A')
       
            if l1 > 7 : l1-=1
            if l2 > 7 : l2-=1
    
            e100km = ((l1-2)%5)*5 +  (l2%5)
            n100km = 19-5*(l1//5) - l2//5
            log.debug("e100k = {}, n100k = {}",e100km,n100km)
            if e100km<0 or e100km>6 or n100km<0 or n100km>12:
                log.error("e100k = {}, n100k = {} - OOR",e100km,n100km)
                raise RuntimeError('Invalid grid reference')
    
            if len(g[2])==0 :
                s=g[1]+g[3]
                c=len(s)//2
                en=[s[:c],s[c:]]
            else:
                en=[g[1],g[3]]
    
            if len(en[0]) != len(en[1]):
                log.error('e=*{}* n=*{}* - unequal lengths',*en)
                raise RuntimeError('Invalid grid reference')
            
            
            en=[int((x+'00000')[:5]) for x in en ]
            log.debug('EN is {}',en)
            factor=pow(10,5)
            parsed= [en[0]+e100km*factor,en[1]+n100km*factor]

        elif isinstance(arg,LatLong):
            if arg.tag!='OSGB36' : 
                log.debug("Transforming {} to OSGB36",arg)
                arg=arg.transform('OSGB36')
                log.debug("Transformed value is {}",arg)
            phi=radians(arg.latitude)
            l  =radians(arg.longitude)
        
            log.debug("phi = {}, l = {}",phi,l)
            v=OSDefaultGrid(phi,l)
            parsed = v.latLongToGrid()
            
        elif len(args)>=2 and areNumbers(args[:2]):
            parsed=args[:2]
        else:
            raise OSGridError('Arguments to OSGridRef constructor must be string or LatLong or E,N')
        
        log.debug('out is {}',parsed)
        self.E, self.N = [floor(x) for x in parsed] # [round(x,0) for x in parsed]
        log.debug("E = {}, N = {}",self.E,self.N)
    

         
    def toLatLong(self,tag='WGS84',viaString=False) :
        if viaString:
            g=OSGridReference(str(self))
            return g.toLatLong(tag=tag,viaString=False)
        
        phi=OSDefaultGrid.phi0
        M=0
        
        log.debug('Starting loop with phi = {}',phi)
        dN=self.N-OSDefaultGrid.N0
        while True:
            phi += (dN-M)/(OSDefaultGrid.aF0())
            M=OSDefaultGrid.meridional(phi)
            if dN-M < 1.0e-5: break
        
        log.debug('Ending loop with phi = {}',phi)
        
        v=OSDefaultGrid(phi,self.E)
        phi,l = v.gridToLatLong()
        
        point =  LatLong(degrees(phi), degrees(l), tag='OSGB36');
        if (tag != 'OSGB36') : 
            point = point.transform(tag)
        return point
    
    def __str__(self):
        try:
            E=int(self.E)
            N=int(self.N)
            e100k = E//100000
            n100k = N//100000
    
            if e100k<0 or e100k>6 or n100k<0 or n100k>12: 
                log.error("e100k = {}, n100k = {} - OOR",e100k,n100k)
                raise OSGridError('Position out of range')
            
            nf=19-n100k
            ef=10+e100k
            
            l1 = nf - (nf%5) + (ef//5)
            l2 = (5*nf)%25 + (ef%5)
            l = [l1,l2]
            log.debug("l = {}",l)
            rect = ''.join([OSGridReference.alphabet[int(x)] for x in l])
            log.debug("l = {}, rect = {}",l,rect)
            
            e = E%100000
            n = N%100000
            log.debug('e = {}, n = {}',e,n)
            
            return ' '.join([rect,str(e).rjust(5,'0'),str(n).rjust(5,'0')])
        except:
            traceback.print_exc()
            log.debug('{}',traceback.format_exc())
            return ''

    def eastNorth(self):
        return (self.E, self.N)
    
    
            
        
 

